/*
	TuxPuck - sprite.cpp
	Copyright (C) 2001 Jacob Kroon

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

	Jacob Kroon - d00jkr@efd.lth.se
*/

/* includes */
#include <string.h>
#include <malloc.h>
#include <SDL_endian.h>
#include <SDL_timer.h>
#include "sprite.h"
#include "global.h"
#include "video.h"

/* Sprite implemention */
Sprite::Sprite(Uint8 * data, Uint32 * memcounter)
{
  Uint32 index = 0;
  Uint32 size;
  Uint8 i;

  memcpy(&size, data, sizeof(Uint32));
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
  size = SDL_Swap32(size);
#endif
  if (memcounter)
    *memcounter += size + sizeof(Uint32);
  data += sizeof(Uint32);
  _nbrOfFrames = data[index++];
  _sdlFrame =
    (SDL_Surface **) calloc(_nbrOfFrames + 1, sizeof(SDL_Surface *));
  for (i = 1; i < _nbrOfFrames + 1; i++)
    _sdlFrame[i] = loadTga(&data[index], &index);
  _nbrOfAnimations = data[index++];
  _animation = (Animation **) calloc(_nbrOfAnimations, sizeof(Animation *));
  for (i = 0; i < _nbrOfAnimations; i++)
    _animation[i] = new Animation(&data[index], &index);
  _x = _y = 0;
  _oldRect.x = _oldRect.y = _oldRect.w = _oldRect.h = 0;
  _currentAnimation = 0;
}

Sprite::~Sprite()
{
  Uint8 i;

  for (i = 1; i < _nbrOfFrames + 1; i++)
    SDL_FreeSurface(_sdlFrame[i]);
  free(_sdlFrame);
  for (i = 0; i < _nbrOfAnimations; i++)
    delete(_animation[i]);
  free(_animation);
}

void Sprite::blit(void)
{
  Uint8 frame;

  frame = _animation[_currentAnimation]->getFrame();
  Video::instance()->erase(&_oldRect);
  if (frame)
    {
      _oldRect.x = _x;
      _oldRect.y = _y;
      Video::instance()->blit(_sdlFrame[frame], &_oldRect);
    }
  else
    _oldRect.x = _oldRect.y = _oldRect.w = _oldRect.h = 0;
}

void Sprite::erase(void)
{
  Video::instance()->erase(&_oldRect);
  _oldRect.x = _oldRect.y = _oldRect.w = _oldRect.h = 0;
}

Uint8 Sprite::getAnimation(void)
{
  return _currentAnimation;
}

void Sprite::setAnimation(Uint8 animation)
{
  _currentAnimation = animation;
  _animation[_currentAnimation]->reset();
}

Uint8 Sprite::update(Uint32 time)
{
  return _animation[_currentAnimation]->update(time);
}

void Sprite::setPosition(Uint32 x, Uint32 y)
{
  _x = x;
  _y = y;
}

/* Animation implemention */
Animation::Animation(Uint8 * data, Uint32 * memcounter)
{
  Uint32 index = 0;
  Uint8 i;

  _nbrOfFrames = data[index++];
  _currentFrame = 0;
  _timeStamp = 0;
  _time = (Uint16 *) calloc(_nbrOfFrames, sizeof(Uint16));
  _frame = (Uint8 *) calloc(_nbrOfFrames, sizeof(Uint8));
  for (i = 0; i < _nbrOfFrames; i++)
    {
      _frame[i] = data[index++];
      memcpy(&_time[i], &data[index], sizeof(Uint16));
      index += 2;
#if SDL_BYTEORDER == SDL_BIG_ENDIAN
      _time[i] = SDL_Swap16(_time[i]);
#endif
    }
  if (memcounter)
    *memcounter += index;
}

Animation::~Animation()
{
  free(_time);
  free(_frame);
}

Uint8 Animation::getFrame(void)
{
  return _frame[_currentFrame];
}

void Animation::reset(void)
{
  _timeStamp = 0;
  _currentFrame = 0;
}

Uint8 Animation::update(Uint32 time)
{
  _timeStamp += time;
  while (_timeStamp > _time[_currentFrame])
    {
      _timeStamp -= _time[_currentFrame];
      _currentFrame++;
      if (_currentFrame > _nbrOfFrames)
	{
	  _currentFrame = 0;
	  return 0;
	}
    }
  return 1;
}
